package FlashDebugger;
use Dancer2;

use threads;
use threads::shared;

use File::Copy;

use File::Basename;
use File::Path qw/make_path/;

use File::Spec;

use Cwd;

use POE qw/Wheel::FollowTail/;

use File::Slurper 'read_text';

$|=1;


our $VERSION = '0.1';

set content_type => 'application/json; charset=UTF-8';
set environment => 'production';

my $encoding : shared;

my $encoding_trace_changed : shared = 0;
my $encoding_policy_changed : shared = 0;

my $max_lines;
my $max_warnings;
my $display_warnings;
my $concat_to_mm;

my $sep = get_separator();
my $home = get_home();

my $trace_log_file = get_flash_app_dir()."${sep}Logs${sep}flashlog.txt";
my $policy_log_file = get_flash_app_dir()."${sep}Logs${sep}policyfiles.txt";

my $flashfirebug_path = File::Spec->canonpath(getcwd()) . "${sep}flashfirebug.swf";
my $security_file = get_flash_app_dir()."${sep}#Security${sep}".
    "FlashPlayerTrust${sep}FlashFirebug.cfg";

my $trace_data : shared = '';
my $policy_data : shared = '';

my $flag_trace : shared =  0;
my $flag_policy : shared =  0;

sub watch_trace_data {
    my $file = shift;

    open(my $fh,'<:encoding('.$encoding.')', $file);

    POE::Session->create(
        inline_states => {
            _start => sub {
                $_[3]{tailor} = POE::Wheel::FollowTail->new(
                    InputEvent => "got_log_line",
                    ResetEvent => "got_log_rollover",
                    Handle => $fh,
                    Seek => 0,
                );
            },
            got_log_line => sub {
                if (index($_[10],
                    'For content targeting Flash Player version 14'.
                        ' or higher, ExternalInterface escapes strings') != -1) {
                    return;
                }


                lock $encoding_trace_changed;

                if ($encoding_trace_changed == 1) {
                    $encoding_trace_changed = 0;
                    binmode($fh, ":encoding($encoding)");
                }

                lock $trace_data;
                $trace_data = $trace_data . $_[10]."\n";
                lock $flag_trace;
                $flag_trace = 1;
            },
            got_log_rollover => sub {

            },
        }
    );

    POE::Kernel->run();
    exit;
}

sub set_mm_cfg {
    ($max_lines, $max_warnings, $display_warnings, $concat_to_mm) = @_;

    if (open (MYFILE, ">${home}${sep}mm.cfg")) {
        my $mm = <<start;
AllowUserLocalTrust=1
AssetCacheSize=20
AutoUpdateDisable=0
AutoUpdateInterval=-1
AVHardwareDisable=0
DisableDeviceFontEnumeration=0
DisableProductDownload=0
ErrorReportingEnable=$display_warnings
FileDownloadDisable=0
FileUploadDisable=0
FullScreenDisable=0
LocalFileReadDisable=0
LocalStorageLimit=6
MaxWarnings=$max_warnings
PolicyFileLog=1
PolicyFileLogAppend=1
PreloadSwf=$flashfirebug_path
TraceOutputFileEnable=1
TraceOutputFileName=$trace_log_file
$concat_to_mm
start

        print MYFILE $mm;
        close (MYFILE);
    }
    else {
        print 'Couldn\'t write to mm.cfg, are you running ',
            'the application with the same user your browser is running in?';
    }

}

sub watch_policy_data {
    my $file = shift;

    open(my $fh,'<:encoding('.$encoding.')', $file);

    POE::Session->create(
        inline_states => {
            _start => sub {
                $_[3]{tailor} = POE::Wheel::FollowTail->new(
                    InputEvent => "got_log_line",
                    ResetEvent => "got_log_rollover",
                    Handle => $fh,
                    Seek => 0,
                );
            },
            got_log_line => sub {
                if (index($_[10],
                    'flashfirebug.swf') != -1) {
                    return;
                }

                lock $encoding_policy_changed;

                if ($encoding_policy_changed == 1) {
                    $encoding_policy_changed = 0;
                    binmode($fh, ":encoding($encoding)");
                }

                lock $policy_data;
                $policy_data = $policy_data . $_[10]."\n";
                lock $flag_policy;
                $flag_policy = 1;
            },
            got_log_rollover => sub {

            },
        }
    );

    POE::Kernel->run();
    exit;
}

sub get_flash_app_dir {
    if ($^O eq 'MSWin32' || $^O eq 'cygwin') {
        return get_app_dir()."${sep}Macromedia${sep}Flash Player";
    }

    if ($^O eq 'darwin') {
        return get_app_dir()."${sep}Preferences${sep}Macromedia${sep}Flash Player";
    }

    # Linux

    return get_app_dir()."${sep}.macromedia${sep}Flash_Player";
}

if ($^O eq 'MSWin32' || $^O eq 'cygwin') {
    eval {
        require Win32::Process;
        Win32::Process->import();
        1;
    } or do {
        my $error = $@;
    };

    Win32::Process::Create(my $ProcessObj,
        File::Spec->canonpath(getcwd()) . "${sep}monitor.exe",
        "monitor $$",
        0,
        0x00000020 | 0x01000000,
        ".")|| die "BAD";
} else {
    $SIG{TERM} = sub {
        if (-f "${home}${sep}mm_backup.cfg") {
            copy("${home}${sep}mm_backup.cfg", "${home}${sep}mm.cfg");
            unlink "${home}${sep}mm_backup.cfg";
        }
    };
}

sub get_app_dir {
    if ($^O eq 'MSWin32' || $^O eq 'cygwin') {
        my $appdata = $ENV{'appdata'};

        if (!defined($appdata)) {
            return File::Spec->canonpath($ENV{'HOME'});
        }

        return File::Spec->canonpath($appdata);
    }

    if ($^O eq 'darwin') { # MAC
        return File::Spec->canonpath($ENV{'HOME'}.$sep.'Library');
    }

    # Linux

    return File::Spec->canonpath($ENV{'HOME'});
}

sub get_separator {
    if ($^O eq 'MSWin32' || $^O eq 'cygwin') {
        return '\\';
    }

    # MAC or linux
    return '/';
}

sub get_home {
    if ($^O eq 'MSWin32' || $^O eq 'cygwin') {
        return $ENV{'HOMEDRIVE'}.File::Spec->canonpath($ENV{'HOMEPATH'});
    }

    # MAC or linux
    return File::Spec->canonpath($ENV{'HOME'});
};

get '/check' => sub {
        return '{"status": "ok"}';
    };

get '/clear/:log' => sub {
        if (route_parameters->get('log') eq 'trace') {
            if (open (MYFILE, ">$trace_log_file")) {
                print MYFILE '';
                close (MYFILE);
            };
        } else {
            if (open (MYFILE, ">$policy_log_file")) {
                print MYFILE '';
                close (MYFILE);
            };
        }

        return '{"status": "ok"}';
    };

post '/start' => sub {
        lock $encoding;
        $encoding = body_parameters->get('encoding');

        $max_lines = body_parameters->get('max_lines');
        $max_warnings = body_parameters->get('max_warnings');
        $display_warnings = body_parameters->get('display_warnings');
        $concat_to_mm = body_parameters->get('concat_to_mm');

        if (open (MYFILE, ">${home}${sep}mm_backup.cfg")) {
            print MYFILE '';
            close (MYFILE);
        };

        if (-f "${home}${sep}mm.cfg") {
            my $content = read_text("${home}${sep}mm.cfg");

            if (index($content, 'flashfirebug.swf') == -1) {
                copy("${home}${sep}mm.cfg", "${home}${sep}mm_backup.cfg")
            }
        }

        set_mm_cfg($max_lines, $max_warnings, $display_warnings, $concat_to_mm);

        make_path(dirname($security_file));
        make_path(dirname($trace_log_file));

        if (open (MYFILE, ">$security_file")) {
            print MYFILE chr(65279).dirname($flashfirebug_path).$sep;
            close (MYFILE);
        };

        if (open (MYFILE, ">$trace_log_file")) {
            print MYFILE '';
            close (MYFILE);
        };

        if (open (MYFILE, ">$policy_log_file")) {
            print MYFILE '';
            close (MYFILE);
        };

        threads->create( \&watch_trace_data, $trace_log_file )->detach;
        threads->create( \&watch_policy_data, $policy_log_file )->detach;

        my %ret = ('status' => 'ok');

        return to_json \%ret;
    };

get '/get-log/:log' => sub {
        if (route_parameters->get('log') eq 'trace') {
            lock $flag_trace;

            my %ret = ('status' => 'ok', 'data' => '');

            # print "BEFORE";
            if ($flag_trace == 1) {
                # print "IN";
                $flag_trace = 0;
                lock $trace_data;
                %ret = ('status' => 'ok', 'data' => $trace_data);
                $trace_data = '';
            }
            # print "AFTER";

            return to_json \%ret;
        }

        lock $flag_policy;

        my %ret = ('status' => 'ok', 'data' => '');

        if ($flag_policy == 1) {
            $flag_policy = 0;
            lock $policy_data;
            %ret = ('status' => 'ok', 'data' => $policy_data);
            $policy_data = '';
        };

        return to_json \%ret;
    };